import os
import gzip
import json
import logging
try:
    from redis import StrictRedis
except ImportError:
    StrictRedis = None

from typing import Any, Dict, List, NamedTuple, Optional, Set, Tuple, Union
from fast_autocomplete import AutoComplete


def read_local_dump(filepath: str):
    with open(filepath, 'r') as the_file:
        return the_file.read()


def _simple_compress(item: str, hash_to_val: Dict[int, str]) -> str:
    item_hash = hash(item)
    if item_hash in hash_to_val:
        item = hash_to_val[item_hash]
    else:
        hash_to_val[item_hash] = item
    return item


class WordValue(NamedTuple):
    context: Any
    display: Any
    count: int = 0
    original_key: 'WordValue' = None

    def get(self, key: str, default: Optional[str] = None) -> str:
        result = getattr(self, key)
        if result is None:
            result = default
        return result


def get_all_content(content_files, redis_client=None, redis_key_prefix=None, logger=None):
    """
    Get all content that is needed to initialize Autocomplete.

    :param: redis_client (optional) If passed, it tries to load from Redis if there is already cached data
    """
    kwargs = {}
    for key, info in content_files.items():
        kwargs[key] = get_data(
            filepath=info['filepath'],
            compress=info['compress'],
            redis_client=redis_client,
            redis_key_prefix=redis_key_prefix,
            logger=logger
        )
    if logger:
        kwargs['logger'] = logger
    return kwargs


def get_data(filepath: str, compress: bool = False,
             redis_client: Optional[StrictRedis] = None,
             redis_key_prefix: Optional[str] = None,
             logger: Optional[logging.RootLogger] = None) -> Dict[str, List[str]]:
    data_json = None
    filename = os.path.basename(filepath)
    if redis_client and redis_key_prefix:
        key = redis_key_prefix.format(filename)
        try:
            data_json = redis_client.get(key)
        except Exception:
            if logger:
                logger.exception('Unable to get the search graph words from Redis.')
            else:
                print('Unable to get the search graph words from Redis.')
        if data_json:
            data_json = gzip.decompress(data_json).decode('utf-8')
    if not data_json:
        data_json = read_local_dump(filepath)
    data = json.loads(data_json)

    if compress:
        hash_to_val = {}

        for word, value in data.items():
            context, display, count = value
            display = _simple_compress(item=display, hash_to_val=hash_to_val)
            for key, val in context.items():
                context[key] = _simple_compress(
                    item=context[key], hash_to_val=hash_to_val
                )
            data[word] = WordValue(context=context, display=display, count=count)

    return data


def populate_redis(content_files, redis_client, redis_cache_prefix):
    """
    Populate Redis with data based on the local files
    """
    for key, info in content_files.items():
        filename = os.path.basename(info['filepath'])
        redis_key = redis_cache_prefix.format(filename)
        data = read_local_dump(info['filepath'])
        compressed = gzip.compress(data.encode('utf-8'))
        redis_client.set(redis_key, compressed)


def autocomplete_factory(
    content_files, redis_client=None, module=AutoComplete, logger=None
):
    """
    Factory function to initialize the proper Vehicle Autocomplete object

    :param: content_files: The file paths and options where data is stored.

    Example

        content_files = {
            'synonyms': {
                'filename': 'path/to/synonyms.json',
                'compress': False
            },
            'words': {
                'filename': 'path/to/words.json',
                'compress': True
            },
            'full_stop_words': {
                'filename': 'path/to/full_stop_words.json',
                'compress': False
            }
        }

    :param: redis_client: (optional) If passed, the factor function tries to load the data from Redis
                                     and if that fails, it will load the local data.
    :param: module: (optional) The AutoComplete module to initialize
    """
    kwargs = get_all_content(content_files, redis_client=redis_client, logger=logger)
    return module(**kwargs)



import io
import os
import csv
import sys
try:
    import termios
    import fcntl
except Exception:
    termios = fcntl = None


class FileNotFound(ValueError):
    pass


def _check_file_exists(path):
    if not os.path.exists(path):
        raise FileNotFound(f'{path} does not exist')


def read_csv_gen(path_or_stringio, csv_func=csv.reader, **kwargs):
    """
    Takes a path_or_stringio to a file or a StringIO object and creates a CSV generator
    """
    if isinstance(path_or_stringio, (str, bytes)):
        _check_file_exists(path_or_stringio)
        encoding = kwargs.pop('encoding', 'utf-8-sig')
        with open(path_or_stringio, 'r', encoding=encoding) as csvfile:
            for i in csv_func(csvfile, **kwargs):
                yield i
    elif isinstance(path_or_stringio, io.StringIO):
        for i in csv_func(path_or_stringio, **kwargs):
            yield i
    else:
        raise TypeError('Either a path to the file or StringIO object needs to be passed.')


def _extend_and_repeat(list1, list2):
    if not list1:
        return [[i] for i in list2]

    result = []
    for item in list2:
        if item not in list1:
            list1_copy = list1.copy()
            if item.startswith(list1_copy[-1]):
                list1_copy.pop()
            list1_copy.append(item)
            result.append(list1_copy)

    return result


def read_single_keypress():
    """Waits for a single keypress on stdin.
    https://stackoverflow.com/a/6599441/1497443

    This is a silly function to call if you need to do it a lot because it has
    to store stdin's current setup, setup stdin for reading single keystrokes
    then read the single keystroke then revert stdin back after reading the
    keystroke.

    Returns the character of the key that was pressed (zero on
    KeyboardInterrupt which can happen when a signal gets handled)

    """
    if fcntl is None or termios is None:
        raise ValueError('termios and/or fcntl packages are not available in your system. This is possible because you are not on a Linux Distro.')
    fd = sys.stdin.fileno()
    # save old state
    flags_save = fcntl.fcntl(fd, fcntl.F_GETFL)
    attrs_save = termios.tcgetattr(fd)
    # make raw - the way to do this comes from the termios(3) man page.
    attrs = list(attrs_save)  # copy the stored version to update
    # iflag
    attrs[0] &= ~(termios.IGNBRK | termios.BRKINT | termios.PARMRK |
                  termios.ISTRIP | termios.INLCR | termios.IGNCR |
                  termios.ICRNL | termios.IXON)
    # oflag
    attrs[1] &= ~termios.OPOST
    # cflag
    attrs[2] &= ~(termios.CSIZE | termios. PARENB)
    attrs[2] |= termios.CS8
    # lflag
    attrs[3] &= ~(termios.ECHONL | termios.ECHO | termios.ICANON |
                  termios.ISIG | termios.IEXTEN)
    termios.tcsetattr(fd, termios.TCSANOW, attrs)
    # turn off non-blocking
    fcntl.fcntl(fd, fcntl.F_SETFL, flags_save & ~os.O_NONBLOCK)
    # read a single keystroke
    try:
        ret = sys.stdin.read(1)  # returns a single character
    except KeyboardInterrupt:
        ret = 0
    finally:
        # restore old state
        termios.tcsetattr(fd, termios.TCSAFLUSH, attrs_save)
        fcntl.fcntl(fd, fcntl.F_SETFL, flags_save)
    return ret



import collections


class DrawGraphMixin:

    DRAW_POPULATION_ANIMATION = False
    DRAW_POPULATION_ANIMATION_PATH = ''
    DRAW_POPULATION_ANIMATION_FILENO_PADDING = 6
    SHOW_OBJ_IDS_OF_WORDS = {}

    def draw_graph(self, file_path, starting_word=None, agraph_kwargs=None, prog='dot'):
        """
        Draws the graph of autocomplete words.

        parameters:

        - file_path: the full path to the file to save the graph into.
        Graphviz library will determine the format of the file based on the extension you choose.
        - starting_word: what word to start from. All descendants of the this word will be in the graph.
        If left as None, the graph will start from the rootn node.
        - agraph_kwargs: kwargs that will be pased to PyGraphViz Agraph creator. You can control how the graph
        will be rendered using these kwargs.
        """
        try:
            import pygraphviz as pgv
        except ImportError:
            print('You need to install pygraphviz in order to draw graphs')

        agraph_kwargs = agraph_kwargs if agraph_kwargs else {}
        graph = pgv.AGraph(strict=False, directed=True, **agraph_kwargs)

        edges = set()
        que = collections.deque()
        if starting_word:
            matched_prefix_of_last_word, rest_of_word, new_node, matched_words = self._prefix_autofill(word=starting_word)
            try:
                matched_word = matched_words[-1]
            except IndexError:
                new_node = self._dwg
                matched_word = 'root'
        else:
            new_node = self._dwg
            matched_word = 'root'
        que.append((matched_word, new_node, ''))
        node_alternative_names = {}
        while que:
            parent_name, node, edge_name = que.popleft()
            node_id = id(node)
            if node_id not in node_alternative_names:
                node_alternative_names[node_id] = f'.{len(node_alternative_names)}'
            if node.word:
                node_name = node.word
                if node_name in self.SHOW_OBJ_IDS_OF_WORDS:
                    node_name = f'{node_name} {id(node)}'
                else:
                    try:
                        node_name = self.words[node_name].display
                    except (KeyError, AttributeError):
                        pass
                graph.add_node(node_name, fontcolor='blue', fontname='Arial', shape='rectangle')
            else:
                node_name = node_alternative_names[node_id]
                graph.add_node(node_name, color='grey', shape='point')
            edge_name = "' '" if edge_name == ' ' else edge_name
            edge = (parent_name, node_name)
            if edge not in edges:
                edges.add(edge)
                graph.add_edge(*edge, color='blue', label=edge_name)
                for edge_name, child in node.children.items():
                    que.append((node_name, child, edge_name))
        graph.draw(file_path, prog=prog)

    def insert_word_callback(self, word):
        """
        Once word is inserted, this call back is run.
        """
        if self.DRAW_POPULATION_ANIMATION:
            if not hasattr(self, '_graph_fileno'):
                self._graph_fileno = 0
                self._graph_filepath = self.DRAW_POPULATION_ANIMATION_PATH.replace('.', r'{}.')

            fileno = str(self._graph_fileno).zfill(self.DRAW_POPULATION_ANIMATION_FILENO_PADDING)
            file_path = self._graph_filepath.format(fileno)
            self.draw_graph(file_path=file_path)
            self._graph_fileno += 1
